"""
Simple row based texture atlas created for fast runtime allocation.

* This atlas is partly based on the texture atlas in the Arcade project
* The allocator is based on Pyglet's row allocator

https://github.com/pyglet/pyglet/blob/master/pyglet/image/atlas.py
https://github.com/pythonarcade/arcade/blob/development/arcade/texture_atlas.py
"""

from typing import Tuple

import moderngl
from .base import BaseImage


class AllocatorException(Exception):
    pass


class _Row:
    """
    A row in the texture atlas.
    """
    __slots__ = ("x", "y", "y2", "max_height")

    def __init__(self, y: int, max_height: int) -> None:
        self.x = 0
        self.y = y
        self.max_height = max_height
        self.y2 = y

    def add(self, width: int, height: int) -> Tuple[int, int]:
        """Add a region to the row and return the position"""
        if width <= 0 or height <= 0:
            raise AllocatorException("Cannot allocate size: [{}, {}]".format(width, height))
        if height > self.max_height:
            raise AllocatorException("Cannot allocate past the max height")

        x, y = self.x, self.y
        self.x += width
        # Keep track of the highest y value for compaction
        self.y2 = max(self.y + height, self.y2)
        return x, y

    def compact(self):
        """
        Compacts the row to the smallest height.
        Should only be done once when the row is filled before adding a new row.
        """
        self.max_height = self.y2 - self.y


class Allocator:
    """Row based allocator"""

    def __init__(self, width: int, height: int):
        self.width = width
        self.height = height
        # First row covers the entire height until compacted
        # when a new row is added
        self.rows = [_Row(0, self.height)]

    def alloc(self, width: int, height: int) -> Tuple[int, int]:
        """
        Allocate a region.

        Returns:
            Tuple[int, int]: The x,y location
        Raises:
            AllocatorException: if no more space
        """
        # Check if we have room in existing rows
        for row in self.rows:
            # Can we add the region to the end if this row?
            if self.width - row.x >= width and row.max_height >= height:
                return row.add(width, height)

        # Can we make a new row?
        if self.width >= width and self.height - row.y2 >= height:
            # Compact the last row
            row.compact()
            # New row continuing from y2 with a remaining of the height as the max
            new_row = _Row(row.y2, self.height - row.y2)
            self.rows.append(new_row)
            # Allocate the are in the new row
            return new_row.add(width, height)

        raise AllocatorException("No more space in {} for box [{}, {}]".format(self, width, height))


class TextureAtlas:
    """
    A simple texture atlas using a row based allocation.
    There are more efficient ways to pack textures, but this
    is normally sufficient for dynamic atlases were textures
    are added on the fly runtime.
    """
    def __init__(
        self,
        ctx: moderngl.Context,
        width: int,
        height: int,
        components: int = 4,
        border: int = 1,
        auto_resize: bool = True,
    ):
        # Basic properties
        self._ctx = ctx
        self._width = width
        self._height = height
        self._components = components
        self._border = border
        self._auto_resize = auto_resize

        # The physical size limit for the current hardware
        self._max_size = self._ctx.info["GL_MAX_VIEWPORT_DIMS"]

        # Atlas content
        self._texture = self._ctx.texture(self.size, components=self._components)

        # We want to be able to render into the atlas texture
        self._fbo = self._fbo = self._ctx.framebuffer(
            color_attachments=[self._texture]
        )
        self._allocator = Allocator(width, height)

    @property
    def ctx(self) -> moderngl.Context:
        """The moderngl contex this atlas belongs to"""
        return self._ctx

    @property
    def textrue(self) -> moderngl.Texture:
        """The moderngl texture with the atlas contents"""
        return self._texture

    @property
    def width(self) -> int:
        """int: Width of the atlas in pixels"""
        return self._width

    @property
    def height(self) -> int:
        """int: Height of the atlas in pixels"""
        return self._height

    @property
    def size(self) -> Tuple[int, int]:
        """Tuple[int, int]: The size of he atlas (width, height)"""
        return self._width, self._height

    @property
    def max_size(self) -> Tuple[int, int]:
        """
        Tuple[int,int]: The maximum size of the atlas in pixels (x, y)
        """
        return self._max_size

    def add(self, image: BaseImage):
        pass

    def remove(self, image: BaseImage):
        pass

    def resize(self, width: int, height: int):
        pass

    def rebuild(self):
        pass



from typing import Tuple
from PIL.Image import Image
from moderngl_window.meta.base import ResourceDescription


class TextureDescription(ResourceDescription):
    """Describes a texture to load.

    Example:

    .. code:: python

        # Loading a 2d texture
        TextureDescription(path='textures/wood.png')

        # Loading a 2d texture with mimpmaps with anisotropy
        TextureDescription(path='textures/wood.png', mipmap=True, anisotropy=16.0)

        # Loading texture array containing 10 layers
        TextureDescription(path='textures/tiles.png', layers=10, kind='array')
    """

    default_kind = "2d"
    resource_type = "textures"

    def __init__(
        self,
        path: str = None,
        kind: str = None,
        flip=True,
        flip_x=False,
        flip_y=True,
        mipmap=False,
        mipmap_levels: Tuple[int, int] = None,
        anisotropy=1.0,
        image=None,
        layers=None,
        pos_x: str = None,
        pos_y: str = None,
        pos_z: str = None,
        neg_x: str = None,
        neg_y: str = None,
        neg_z: str = None,
        **kwargs
    ):
        """Describes a texture resource

        Args:
            path (str): path to resource relative to search directories
            kind (str): The kind of loader to use
            flip (boolean): (use flip_y) Flip the image vertically (top to bottom)
            flip_x (boolean): Flip the image horizontally (left to right)
            flip_y (boolean): Flip the image vertically (top to bottom)
            mipmap (bool): Generate mipmaps. Will generate max possible levels unless
                           `mipmap_levels` is defined.
            mipmap_levels (tuple): (base, max_level) controlling mipmap generation.
                                   When defined the `mipmap` parameter is automatically `True`.
            anisotropy (float): Number of samples for anisotropic filtering
            image: PIL image for when loading embedded resources
            layers: (int): Number of layers for texture arrays
            neg_x (str): Path to negative x texture in a cube map
            neg_y (str): Path to negative y texture in a cube map
            neg_z (str): Path to negative z texture in a cube map
            pos_x (str): Path to positive x texture in a cube map
            pop_y (str): Path to positive y texture in a cube map
            pos_z (str): Path to positive z texture in a cube map
            **kwargs: Any optional/custom attributes
        """
        kwargs.update(
            {
                "path": path,
                "kind": kind,
                "flip_x": flip_x,
                "flip_y": flip and flip_y,
                "mipmap": mipmap,
                "mipmap_levels": mipmap_levels,
                "anisotropy": anisotropy,
                "layers": layers,
                "image": image,
                "neg_x": neg_x,
                "neg_y": neg_y,
                "neg_z": neg_z,
                "pos_x": pos_x,
                "pos_y": pos_y,
                "pos_z": pos_z,
            }
        )
        super().__init__(**kwargs)

    @property
    def flip_x(self) -> bool:
        """bool: If the image should be flipped horizontally (left to right)"""
        return self._kwargs.get("flip_x")

    @property
    def flip_y(self) -> bool:
        """bool: If the image should be flipped vertically (top to bottom)"""
        return self._kwargs.get("flip_y")

    @property
    def mipmap(self) -> bool:
        """bool: If mipmaps should be generated"""
        return self._kwargs.get("mipmap")

    @mipmap.setter
    def mipmap(self, value: float):
        self._kwargs["mipmap"] = value

    @property
    def mipmap_levels(self) -> Tuple[int, int]:
        """Tuple[int, int]: base, max_level for mipmap generation"""
        return self._kwargs.get("mipmap_levels")

    @property
    def layers(self) -> int:
        """int: Number of layers in texture array"""
        return self._kwargs.get("layers")

    @property
    def anisotropy(self) -> float:
        """float: Number of samples for anisotropic filtering"""
        return self._kwargs.get("anisotropy")

    @property
    def image(self) -> Image:
        """Image: PIL image when loading embedded resources"""
        return self._kwargs.get("image")

    @property
    def pos_x(self):
        """str: Path to positive x in a cubemap texture"""
        return self._kwargs.get("pos_x")

    @property
    def pos_y(self):
        """str: Path to positive y in a cubemap texture"""
        return self._kwargs.get("pos_y")

    @property
    def pos_z(self):
        """str: Path to positive z in a cubemap texture"""
        return self._kwargs.get("pos_z")

    @property
    def neg_x(self):
        """str: Path to negative x in a cubemap texture"""
        return self._kwargs.get("neg_x")

    @property
    def neg_y(self):
        """str: Path to negative y in a cubemap texture"""
        return self._kwargs.get("neg_y")

    @property
    def neg_z(self):
        """str: Path to negative z in a cubemap texture"""
        return self._kwargs.get("neg_z")



from PIL.Image import Image
from typing import Any, Tuple


class BaseImage:
    """
    Base class for atlas images.
    Must be hashable for use as dict keys and in sets
    and decides the uniqueness of the image.
    """

    def __init__(self, image: Any):
        self._image = image

    @property
    def width(self) -> int:
        """int: Width of the image in pixels"""
        raise NotImplementedError

    @property
    def height(self) -> int:
        """int: Height of the image in pixels"""
        raise NotImplementedError

    @property
    def size(self) -> Tuple[int, int]:
        """Tuple[int, int]: Size of the image in pixels (width, height)"""
        raise NotImplementedError

    def get_pixel_data(self, components: int = 4):
        """
        Get the raw pixel data from the image.

        Keyword Args:
            components: Number of components to get
        """
        raise NotImplementedError

    def __hash__(self) -> int:
        return id(self)


class AtlasImage(BaseImage):
    """An atlas image using Pillow"""

    def __init__(self, image: Image):
        self.image = image

    @property
    def width(self) -> int:
        return self._image.size[0]

    @property
    def height(self) -> int:
        return self._image.size[1]

    @property
    def size(self) -> Tuple[int, int]:
        return self._image.size

    def get_pixel_data(self, components: int = 4):
        """
        Get the raw pixel data from the image.

        Keyword Args:
            components: Number of components to get
        """
        if components == 4:
            return self._image.convert("RGBA").tobytes()
        elif components == 3:
            return self._image.covert("RGB").tobytes()
        else:
            raise ValueError("Only supports 3 or 4 components")



# flake8: noqa E741
from moderngl_window.context.base import BaseKeys


class Keys(BaseKeys):
    """
    Namespace mapping tkinter keys.
    Maps the keysym strings provided in tk.Events
    """

    # TODO: We might want to use keycode instead if this works cross platform.
    #       Right now upper case characters will not be recognized.
    #
    # Possibly we need to separate actual characters and control keys.
    #
    # <KeyPress event state=Mod1 keysym=d keycode=68 char='d'>
    # <KeyPress event state=Shift|Mod1 keysym=D keycode=68 char='D'>
    #
    # See : https://www.tcl.tk/man/tcl8.4/TkCmd/keysyms.htm
    ESCAPE = "Escape"
    SPACE = " "
    ENTER = "Return"
    PAGE_UP = "Prior"
    PAGE_DOWN = "Next"
    LEFT = "Left"
    RIGHT = "Right"
    UP = "Up"
    DOWN = "Down"

    TAB = "Tab"
    COMMA = "comma"
    MINUS = "minus"
    PERIOD = "period"
    SLASH = "slash"
    SEMICOLON = "semicolon"
    EQUAL = "equal"
    LEFT_BRACKET = "bracketleft"
    RIGHT_BRACKET = "bracketright"
    BACKSLASH = "backslash"
    BACKSPACE = "BackSpace"
    INSERT = "Insert"
    DELETE = "Delete"
    HOME = "Home"
    END = "End"
    CAPS_LOCK = "Caps_Lock"

    F1 = "F1"
    F2 = "F2"
    F3 = "F3"
    F4 = "F4"
    F5 = "F5"
    F6 = "F6"
    F7 = "F7"
    F8 = "F8"
    F9 = "F9"
    F10 = "F10"
    F11 = "F11"
    F12 = "F12"

    NUMBER_0 = "0"
    NUMBER_1 = "1"
    NUMBER_2 = "2"
    NUMBER_3 = "3"
    NUMBER_4 = "4"
    NUMBER_5 = "5"
    NUMBER_6 = "6"
    NUMBER_7 = "7"
    NUMBER_8 = "8"
    NUMBER_9 = "9"

    NUMPAD_0 = "0"
    NUMPAD_1 = "1"
    NUMPAD_2 = "2"
    NUMPAD_3 = "3"
    NUMPAD_4 = "4"
    NUMPAD_5 = "5"
    NUMPAD_6 = "6"
    NUMPAD_7 = "7"
    NUMPAD_8 = "8"
    NUMPAD_9 = "9"

    A = "a"
    B = "b"
    C = "c"
    D = "d"
    E = "e"
    F = "f"
    G = "g"
    H = "h"
    I = "i"
    J = "j"
    K = "k"
    L = "l"
    M = "m"
    N = "n"
    O = "o"
    P = "p"
    Q = "q"
    R = "r"
    S = "s"
    T = "t"
    U = "u"
    V = "v"
    W = "w"
    X = "x"
    Y = "y"
    Z = "z"



import os

import numpy
import moderngl
import moderngl_window

from moderngl_window.conf import settings
from moderngl_window.resources import programs
from moderngl_window.meta import ProgramDescription


settings.PROGRAM_DIRS.append(os.path.join(os.path.dirname(__file__), "programs"))


class MeshProgram:
    """
    Describes how a mesh is rendered using a specific shader program
    """

    def __init__(self, program: moderngl.Program = None, **kwargs):
        """Initialize.

        Args:
            program: The moderngl program
        """
        self.program = program

    @property
    def ctx(self):
        """moderngl.Context: The current context"""
        return moderngl_window.ctx()

    def draw(
        self,
        mesh,
        projection_matrix: numpy.ndarray = None,
        model_matrix: numpy.ndarray = None,
        camera_matrix: numpy.ndarray = None,
        time=0.0,
    ):
        """Draw code for the mesh

        Args:
            mesh (Mesh): The mesh to render
        Keyword Args:
            projection_matrix (numpy.ndarray): projection_matrix (bytes)
            model_matrix (numpy.ndarray): view_matrix (bytes)
            camera_matrix (numpy.ndarray): camera_matrix (bytes)
            time (float): The current time
        """
        self.program["m_proj"].write(projection_matrix)
        self.program["m_mv"].write(model_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh):
        """
        Determine if this ``MeshProgram`` should be applied to the mesh.
        Can return self or some ``MeshProgram`` instance to support dynamic ``MeshProgram`` creation

        Args:
            mesh: The mesh to inspect
        """
        raise NotImplementedError(
            "apply is not implemented. Please override the MeshProgram method"
        )


class VertexColorProgram(MeshProgram):
    """Vertex color program"""

    def __init__(self, program=None, **kwargs):
        super().__init__(program=None)
        self.program = programs.load(
            ProgramDescription(path="scene_default/vertex_color.glsl")
        )

    def draw(
        self,
        mesh,
        projection_matrix=None,
        model_matrix=None,
        camera_matrix=None,
        time=0,
    ):
        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh):
        if not mesh.material:
            return None

        if mesh.attributes.get("TEXCOORD_0"):
            return None

        if mesh.attributes.get("COLOR_0"):
            return self

        return None


class ColorLightProgram(MeshProgram):
    """Simple color program with light"""

    def __init__(self, program=None, **kwargs):
        super().__init__(program=None)
        self.program = programs.load(
            ProgramDescription(path="scene_default/color_light.glsl")
        )

    def draw(
        self,
        mesh,
        projection_matrix=None,
        model_matrix=None,
        camera_matrix=None,
        time=0,
    ):

        if mesh.material:
            # if mesh.material.double_sided:
            #     self.ctx.disable(moderngl.CULL_FACE)
            # else:
            #     self.ctx.enable(moderngl.CULL_FACE)

            if mesh.material.color:
                self.program["color"].value = tuple(mesh.material.color)
            else:
                self.program["color"].value = (1.0, 1.0, 1.0, 1.0)

        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh):
        if not mesh.material:
            return None

        if not mesh.attributes.get("NORMAL"):
            return None

        return self


class TextureProgram(MeshProgram):
    """Plan textured"""

    def __init__(self, program=None, **kwargs):
        super().__init__(program=None)
        self.program = programs.load(
            ProgramDescription(path="scene_default/texture.glsl")
        )

    def draw(
        self,
        mesh,
        projection_matrix=None,
        model_matrix=None,
        camera_matrix=None,
        time=0,
    ):
        mesh.material.mat_texture.texture.use()
        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh):
        if not mesh.material:
            return None

        if mesh.attributes.get("NORMAL"):
            return None

        if not mesh.attributes.get("TEXCOORD_0"):
            return None

        if mesh.attributes.get("COLOR_0"):
            return None

        if mesh.material.mat_texture is not None:
            return self

        return None


class TextureVertexColorProgram(MeshProgram):
    """textred object with vertex color"""

    def __init__(self, program=None, **kwargs):
        super().__init__(program=None)
        self.program = programs.load(
            ProgramDescription(path="scene_default/vertex_color_texture.glsl")
        )

    def draw(
        self,
        mesh,
        projection_matrix=None,
        model_matrix=None,
        camera_matrix=None,
        time=0,
    ):
        mesh.material.mat_texture.texture.use()
        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh):
        if not mesh.material:
            return None

        if mesh.attributes.get("NORMAL"):
            return None

        if not mesh.attributes.get("TEXCOORD_0"):
            return None

        if not mesh.attributes.get("COLOR_0"):
            return None

        if mesh.material.mat_texture is not None:
            return self

        return None


class TextureLightProgram(MeshProgram):
    """
    Simple texture program
    """

    def __init__(self, program=None, **kwargs):
        super().__init__(program=None)
        self.program = programs.load(
            ProgramDescription(path="scene_default/texture_light.glsl")
        )

    def draw(
        self,
        mesh,
        projection_matrix=None,
        model_matrix=None,
        camera_matrix=None,
        time=0,
    ):
        # if mesh.material.double_sided:
        #     self.ctx.disable(moderngl.CULL_FACE)
        # else:
        #     self.ctx.enable(moderngl.CULL_FACE)

        mesh.material.mat_texture.texture.use()
        self.program["texture0"].value = 0
        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)
        mesh.vao.render(self.program)

    def apply(self, mesh):
        if not mesh.material:
            return None

        if not mesh.attributes.get("NORMAL"):
            return None

        if not mesh.attributes.get("TEXCOORD_0"):
            return None

        if mesh.material.mat_texture is not None:
            return self

        return None


class TextureLightColorProgram:
    pass


class FallbackProgram(MeshProgram):
    """
    Fallback program only rendering positions in white
    """

    def __init__(self, program=None, **kwargs):
        super().__init__(program=None)
        self.program = programs.load(
            ProgramDescription(path="scene_default/fallback.glsl")
        )

    def draw(
        self,
        mesh,
        projection_matrix=None,
        model_matrix=None,
        camera_matrix=None,
        time=0,
    ):

        self.program["m_proj"].write(projection_matrix)
        self.program["m_model"].write(model_matrix)
        self.program["m_cam"].write(camera_matrix)

        if mesh.material:
            self.program["color"].value = tuple(mesh.material.color[0:3])
        else:
            self.program["color"].value = (1.0, 1.0, 1.0)

        mesh.vao.render(self.program)

    def apply(self, mesh):
        return self



from pathlib import Path
from typing import List, Union

from contextlib import contextmanager
from moderngl_window.conf import settings


@contextmanager
def texture_dirs(paths: List[Union[Path, str]]):
    """Context manager temporarily replacing texture paths
    Args:
        paths (List[Union[Path, str]]): list of paths
    """
    original_dirs = settings.DATA_DIRS
    settings.TEXTURE_DIRS = paths
    yield None
    settings.TEXTURE_DIRS = original_dirs



# pylint: disable = missing-docstring
from .scene import Scene  # noqa
from .node import Node  # noqa
from .camera import Camera  # noqa
from .camera import KeyboardCamera  # noqa
from .mesh import Mesh  # noqa
from .material import Material, MaterialTexture  # noqa
from .programs import MeshProgram  # noqa



import moderngl
from pathlib import Path
import moderngl_window as mglw
from moderngl_window import geometry
from pyrr import Matrix44

# from moderngl_window.conf import settings
# settings.SCREENSHOT_PATH = 'screenshots'
# from moderngl_window import screenshot


class Test(mglw.WindowConfig):
    title = "Animated Sprite"
    resource_dir = (Path(__file__) / '../../resources').resolve()
    aspect_ratio = 320 / 256
    window_size = 320 * 3, 256 * 3

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.buffer_size = 320, 256
        # Textures
        self.background_texture = self.load_texture_array('textures/animated_sprites/giphy.gif')
        self.background_texture.repeat_x = False
        self.background_texture.repeat_y = False
        self.caveman_texture = self.load_texture_array('textures/animated_sprites/player_2.gif', layers=35)
        self.caveman_texture.repeat_x = False
        self.caveman_texture.repeat_y = False
        self.caveman_texture.filter = moderngl.NEAREST, moderngl.NEAREST

        # Geometry
        # One pixel quad 0, 0 -> 1.0, 1.0
        self.sprite_geometry = geometry.quad_2d(size=(1.0, 1.0), pos=(0.5, 0.5))
        self.quad_fs = geometry.quad_fs()

        # Programs
        self.sprite_program = self.load_program('programs/animated_sprites/sprite_array.glsl')
        self.texture_program = self.load_program('programs/texture.glsl')

        # Offscreen buffer
        self.offscreen_texture = self.ctx.texture(self.buffer_size, 4)
        self.offscreen_texture.filter = moderngl.NEAREST, moderngl.NEAREST
        self.offscreen = self.ctx.framebuffer(color_attachments=[self.offscreen_texture])

        self.projection = Matrix44.orthogonal_projection(0, 320, 0, 256, -1.0, 1.0, dtype='f4')
        self.sprite_program['projection'].write(self.projection)

    def render(self, time, frame_time):
        # Render sprite of offscreen
        self.offscreen.use()
        self.ctx.clear(0.5, 0.5, 0.5, 0.0)

        self.render_sprite(self.background_texture, frame=int(time * 15) % self.background_texture.layers)
        self.render_sprite(self.caveman_texture, frame=int(time * 15) % self.caveman_texture.layers, blend=True, position=(260, 20))

        # Display offscreen
        self.ctx.screen.use()
        self.offscreen_texture.use(location=0)
        self.quad_fs.render(self.texture_program)

        # if self.wnd.frames < 100:
        #     screenshot.create(self.ctx.screen)

    def render_sprite(self, texture, blend=False, frame=0, position=(0, 0)):
        if blend:
            self.ctx.enable(moderngl.BLEND)

        texture.use(location=0)
        self.sprite_program['layer_id'] = frame
        self.sprite_program['position'] = position
        self.sprite_geometry.render(self.sprite_program)

        if blend:
            self.ctx.disable(moderngl.BLEND)


if __name__ == '__main__':
    mglw.run_window_config(Test)
    # Test.run()



from pathlib import Path
import numpy as np
from pyrr import Matrix44

import moderngl
import moderngl_window
from moderngl_window.opengl.vao import VAO
from moderngl_window import geometry
from base import CameraWindow


class VolumetricTetrahedralMesh(CameraWindow):
    """Volumetric Tetrahedral Mesh.

    The dataset was provided by:
    Mara Catalina Aguilera Canon at the Bournemouth University (UK).
    Area of research: Graph Neuro Networks, Finite Element Method

    An example rendering a volumetric mesh of the format:
    ``[[p1, p2, p3, p4], [p1, p2, p3, p4], ..]``
    were ```px``` represent a 3d point in a tetraherdon.
    A geometry shader calculates and emits the tetraherdons
    as triangles and calculate normals on the fly while rendering data.

    This helps us avoid doing this expensive operation
    in python and greatly reduces the memory requirement.

    Controls:
    - Camera: Mouse for rotation. AWSD + QE for translation
    - Press b to toggle blend mode on/off
    - Mouse wheel to increase or decrease the threshold for a tetra to be alive
    """
    gl_version = (4, 1)
    title = "Volumetric Tetrahedra lMesh"
    aspect_ratio = None
    resource_dir = (Path(__file__) / '../../resources').resolve()
    samples = 4

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Finetune camera
        self.wnd.mouse_exclusivity = True
        self.camera.projection.update(near=.01, far=100)
        self.camera.mouse_sensitivity = .5
        self.camera.velocity = 2.5
        self.camera.projection.update(fov=60)

        # Scene states
        self.with_blending = False
        self.line_color = (0.0, 0.0, 0.0)
        self.mesh_color = (0.0, 0.8, 0.0)
        self.threshold = 0.5

        # For rendering background
        self.quad_fs = geometry.quad_fs()

        # (172575,) | 57,525 vertices
        vertices = np.load(self.resource_dir / 'data/tetrahedral_mesh/mesh_nodes.npy')
        vertices = np.concatenate(vertices)
        # (259490, 4) (1037960,) indices
        indices = np.load(self.resource_dir / 'data/tetrahedral_mesh/element_nodes.npy')
        indices = np.concatenate(indices) - 1

        # Probability of a tetrahedron is still alive
        w, h = 8192, int(np.ceil(indices.shape[0] / 8192))
        self.alive_data = np.random.random_sample(w * h)
        self.alive_texture = self.ctx.texture((w, h), 1, dtype='f2')
        self.alive_texture.write(self.alive_data.astype('f2'))

        # Original geometry with indices
        self.geometry = VAO(name='geometry_indices')
        self.geometry.buffer(vertices, '3f', 'in_position')
        self.geometry.index_buffer(indices, index_element_size=4)

        self.prog_background = self.load_program('programs/tetrahedral_mesh/bg.glsl')
        self.prog_gen_tetra = self.load_program(
            vertex_shader='programs/tetrahedral_mesh/gen_tetra_vert.glsl',
            geometry_shader='programs/tetrahedral_mesh/gen_tetra_geo.glsl',
            fragment_shader='programs/tetrahedral_mesh/gen_tetra_frag.glsl',
        )
        self.prog_gen_tetra_lines = self.load_program(
            vertex_shader='programs/tetrahedral_mesh/gen_tetra_vert.glsl',
            geometry_shader='programs/tetrahedral_mesh/gen_tetra_geo.glsl',
            fragment_shader='programs/tetrahedral_mesh/lines_frag.glsl',
        )

        # Query object for measuring the rendering call in OpenGL
        # It delivers the GPU time it took to process commands
        self.query = self.ctx.query(samples=True, any_samples=True, time=True, primitives=True)
        self.total_elapsed = 0

    def render(self, time, frametime):

        # Render background
        self.ctx.wireframe = False
        if not self.with_blending:
            self.ctx.enable_only(moderngl.NOTHING)
            self.quad_fs.render(self.prog_background)

        # Handle blend mode toggle
        if self.with_blending:
            self.ctx.enable_only(moderngl.BLEND)
            self.ctx.blend_func = moderngl.ONE, moderngl.ONE
        else:
            self.ctx.enable_only(moderngl.DEPTH_TEST | moderngl.CULL_FACE)

        # Render tetrahedral mesh
        translate = Matrix44.from_translation((0.0, 2.5, -15.0), dtype='f4')
        rotate = Matrix44.from_eulers((np.radians(180), 0, 0), dtype='f4')
        scale = Matrix44.from_scale((400, 400, 400), dtype='f4')
        mat = self.camera.matrix * translate * rotate * scale

        # All render calls inside this context are timed
        with self.query:
            self.alive_texture.use(location=0)
            self.prog_gen_tetra['alive_texture'].value = 0
            self.prog_gen_tetra['threshold'].value = self.threshold
            self.prog_gen_tetra['color'].value = self.mesh_color
            self.prog_gen_tetra['m_cam'].write(mat)
            self.prog_gen_tetra['m_proj'].write(self.camera.projection.matrix)
            self.geometry.render(self.prog_gen_tetra, mode=moderngl.LINES_ADJACENCY)

            # Render lines
            self.ctx.wireframe = True
            self.alive_texture.use(location=0)
            self.prog_gen_tetra_lines['alive_texture'].value = 0
            self.prog_gen_tetra_lines['threshold'].value = self.threshold
            self.prog_gen_tetra_lines['color'].value = self.line_color
            self.prog_gen_tetra_lines['m_cam'].write(mat)
            self.prog_gen_tetra_lines['m_proj'].write(self.camera.projection.matrix)
            self.geometry.render(self.prog_gen_tetra_lines, mode=moderngl.LINES_ADJACENCY)

        self.total_elapsed = self.query.elapsed

    def key_event(self, key, action, modifiers):
        super().key_event(key, action, modifiers)
        keys = self.wnd.keys

        if action == keys.ACTION_PRESS:
            if key == keys.B:
                self.with_blending = not self.with_blending
                print("With blending:", self.with_blending)
                if self.with_blending:
                    self.mesh_color = 0.01, 0.01, 0.01
                    self.line_color = 0.01, 0.01, 0.01
                else:
                    self.mesh_color = 0.0, 0.8, 0.0
                    self.line_color = 0.0, 0.0, 0.0

    def mouse_scroll_event(self, x_offset, y_offset):
        if y_offset > 0:
            self.threshold += 0.01
        else:
            self.threshold -= 0.01

        self.threshold = max(min(self.threshold, 1.0), 0.0)

    def close(self):
        # 1 s = 1000000000 ns
        # 1 s = 1000000 μs
        avg = self.total_elapsed / self.wnd.frames
        print("Average rendering time per frame: {} ns | {} μs".format(
            round(avg, 4),  # ns
            round(avg / 1000, 4),  # μs
        ))


if __name__ == '__main__':
    VolumetricTetrahedralMesh.run()



import moderngl
import numpy

from headless import HeadlessTestCase
from moderngl_window.opengl.vao import VAO, BufferInfo, VAOError


class VaoTestCase(HeadlessTestCase):

    def createProgram(self):
        return self.ctx.program(
            vertex_shader="""
            #version 330

            in vec3 position;
            in vec3 normal;
            out vec3 v_normal;

            void main() {
                gl_Position = vec4(position, 1.0);
                v_normal = normalize(normal);
            }
            """,
            fragment_shader="""
            #version 330

            out vec4 fragColor;
            in vec3 v_normal;

            void main() {
                float col = length(v_normal);
                fragColor = vec4(col);
            }
            """,
        )

    def test_create(self):
        mesh = VAO("test", mode=moderngl.LINES)
        mesh.buffer(numpy.array([0.0, 0.0, 0.0, 1.0, 1.0, 1.0], dtype='f4'), '3f', 'position')
        mesh.buffer(numpy.array([0.0, 0.0, 1.0, 1.0, 0.0, 1.0], dtype='f4'), '3f', 'normal')
        mesh.buffer(numpy.array([0.0, 0.0, 1.0, 1.0], dtype='f4'), '2f', 'uv')

        # Ensure basic properties are correct
        self.assertEqual(mesh.name, "test")
        self.assertEqual(mesh.vertex_count, 2)
        self.assertEqual(mesh.mode, moderngl.LINES)

        # Ensure buffers are present
        self.assertIsInstance(mesh.get_buffer_by_name('position'), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name('normal'), BufferInfo)
        self.assertIsInstance(mesh.get_buffer_by_name('uv'), BufferInfo)
        self.assertEqual(mesh.get_buffer_by_name('something'), None)

        # Create a progam using a subset of the buffers
        prog = self.createProgram()
        vao = mesh.instance(prog)
        self.assertIsInstance(vao, moderngl.VertexArray)

        # Render directly with VAO and VertexArray instance
        mesh.render(prog)
        vao.render()

    def test_illegal_draw_mode(self):
        """Create vao with illegal draw mode"""
        with self.assertRaises(VAOError):
            VAO(mode=1337)

    def test_add_illegal_buffer(self):
        """Attempt to add illegal buffer"""
        vao = VAO()
        with self.assertRaises(VAOError):
            vao.buffer("stuff", '1f', 'in_position')

    def test_normalized_types(self):
        """Ensure VAO wrapper can handle normalized types"""
        prog = self.createProgram()
        mesh = VAO("test", mode=moderngl.LINES)
        mesh.buffer(numpy.array([0.0, 0.0, 0.0, 1.0, 1.0, 1.0], dtype='i4'), '3ni', 'position')
        mesh.buffer(numpy.array([0.0, 0.0, 1.0, 1.0, 0.0, 1.0], dtype='f4'), '3nf', 'normal')
        mesh.buffer(numpy.array([0.0, 0.0, 1.0, 1.0], dtype='f4'), '2f', 'uv')

        # Uncomment in moderngl 5.6+
        # mesh.instance(prog)

    def test_divisors(self):
        """Test defining buffers with different divisor types"""
        mesh = VAO("test", mode=moderngl.LINES)
        mesh.buffer(numpy.array([0.0, 0.0, 0.0, 1.0, 1.0, 1.0], dtype='f4'), '3f/v', 'position')
        mesh.buffer(numpy.array([0.0, 0.0, 1.0, 1.0, 0.0, 1.0], dtype='f4'), '3f/r', 'normal')
        mesh.buffer(numpy.array([0.0, 0.0, 1.0, 1.0], dtype='f4'), '2f/i', 'uv')

        buffer1 = mesh.get_buffer_by_name('position')
        buffer2 = mesh.get_buffer_by_name('normal')
        buffer3 = mesh.get_buffer_by_name('uv')

        attributes = ['position', 'normal', 'uv']
        self.assertEqual(buffer1.content(attributes), (buffer1.buffer, '3f/v', 'position'))
        self.assertEqual(buffer2.content(attributes), (buffer2.buffer, '3f/r', 'normal'))
        self.assertEqual(buffer3.content(attributes), (buffer3.buffer, '2f/i', 'uv'))



from pyrr import matrix44
import numpy


class Mesh:
    """Mesh info and geometry"""

    def __init__(
        self,
        name,
        vao=None,
        material=None,
        attributes=None,
        bbox_min=None,
        bbox_max=None,
    ):
        """Initialize mesh.

        Args:
            name (str): name of the mesh
        Keyword Args:
            vao (VAO): geometry
            material (Material): material for the mesh
            attributes (dict): Details info about each mesh attribute (dict)
            bbox_min: xyz min values
            bbox_max: xyz max values

        Attributes example::

            {
                "NORMAL": {"name": "in_normal", "components": 3, "type": GL_FLOAT},
                "POSITION": {"name": "in_position", "components": 3, "type": GL_FLOAT}
            }
        """
        self.name = name
        self.vao = vao
        self.material = material
        self.attributes = attributes or {}
        self.bbox_min = bbox_min
        self.bbox_max = bbox_max
        self.mesh_program = None

    def draw(
        self, projection_matrix=None, model_matrix=None, camera_matrix=None, time=0.0
    ):
        """Draw the mesh using the assigned mesh program

        Keyword Args:
            projection_matrix (bytes): projection_matrix
            view_matrix (bytes): view_matrix
            camera_matrix (bytes): camera_matrix
        """
        if self.mesh_program:
            self.mesh_program.draw(
                self,
                projection_matrix=projection_matrix,
                model_matrix=model_matrix,
                camera_matrix=camera_matrix,
                time=time,
            )

    def draw_bbox(self, proj_matrix, model_matrix, cam_matrix, program, vao):
        """Renders the bounding box for this mesh.

        Args:
            proj_matrix: Projection matrix
            model_matrix: View/model matrix
            cam_matrix: Camera matrix
            program: The moderngl.Program rendering the bounding box
            vao: The vao mesh for the bounding box
        """
        program["m_proj"].write(proj_matrix)
        program["m_model"].write(model_matrix)
        program["m_cam"].write(cam_matrix)
        program["bb_min"].write(self.bbox_min.astype("f4").tobytes())
        program["bb_max"].write(self.bbox_max.astype("f4").tobytes())
        vao.render(program)

    def draw_wireframe(self, proj_matrix, model_matrix, program):
        """Render the mesh as wireframe.

            proj_matrix: Projection matrix
            model_matrix: View/model matrix
            program: The moderngl.Program rendering the wireframe
        """
        program["m_proj"].write(proj_matrix)
        program["m_model"].write(model_matrix)
        self.vao.render(program)

    def add_attribute(self, attr_type, name, components):
        """
        Add metadata about the mesh
        :param attr_type: POSITION, NORMAL etc
        :param name: The attribute name used in the program
        :param components: Number of floats
        """
        self.attributes[attr_type] = {"name": name, "components": components}

    def calc_global_bbox(self, view_matrix, bbox_min, bbox_max):
        """Calculates the global bounding.

        Args:
            view_matrix: View matrix
            bbox_min: xyz min
            bbox_max: xyz max
        Returns:
            bbox_min, bbox_max: Combined bbox
        """
        # Copy and extend to vec4
        bb1 = numpy.append(self.bbox_min[:], 1.0).astype("f4")
        bb2 = numpy.append(self.bbox_max[:], 1.0).astype("f4")

        # Transform the bbox values
        bmin = (matrix44.apply_to_vector(view_matrix, bb1),)
        bmax = (matrix44.apply_to_vector(view_matrix, bb2),)
        bmin = numpy.asarray(bmin, dtype="f4")[0]
        bmax = numpy.asarray(bmax, dtype="f4")[0]

        # If a rotation happened there is an axis change and we have to ensure max-min is positive
        for i in range(3):
            if bmax[i] - bmin[i] < 0:
                bmin[i], bmax[i] = bmax[i], bmin[i]

        if bbox_min is None or bbox_max is None:
            return bmin[0:3], bmax[0:3]

        for i in range(3):
            bbox_min[i] = min(bbox_min[i], bmin[i])

        for i in range(3):
            bbox_max[i] = max(bbox_max[i], bmax[i])

        return bbox_min, bbox_max

    def has_normals(self) -> bool:
        """
        Returns:
            bool: Does the mesh have a normals?
        """
        return "NORMAL" in self.attributes

    def has_uvs(self, layer=0) -> bool:
        """
        Returns:
            bool: Does the mesh have texture coordinates?
        """
        return "TEXCOORD_{}".format(layer) in self.attributes



from moderngl_window.geometry.attributes import AttributeNames  # noqa
from moderngl_window.geometry.cube import cube  # noqa
from moderngl_window.geometry.bbox import bbox  # noqa
from moderngl_window.geometry.sphere import sphere  # noqa
from moderngl_window.geometry.quad import quad_2d, quad_fs  # noqa



from pathlib import Path
from pyrr import Matrix44, Vector3

import moderngl
import moderngl_window
from moderngl_window import geometry

from base import CameraWindow


class CubeSimple(CameraWindow):
    title = "Plain Cube"
    resource_dir = (Path(__file__).parent / 'resources').resolve()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.wnd.mouse_exclusivity = True
        self.cube = geometry.cube(size=(2, 2, 2))
        self.prog = self.load_program('programs/cube_simple.glsl')
        self.prog['color'].value = 1.0, 1.0, 1.0, 1.0

    def render(self, time: float, frametime: float):
        self.ctx.enable_only(moderngl.CULL_FACE | moderngl.DEPTH_TEST)

        rotation = Matrix44.from_eulers((time, time, time), dtype='f4')
        translation = Matrix44.from_translation((0.0, 0.0, -3.5), dtype='f4')
        modelview = translation * rotation

        self.prog['m_proj'].write(self.camera.projection.matrix)
        self.prog['m_model'].write(modelview)
        self.prog['m_camera'].write(self.camera.matrix)

        self.cube.render(self.prog)


if __name__ == '__main__':
    moderngl_window.run_window_config(CubeSimple)
